/*
 * Copyright (C) 2016 Moritz Orbach <zufall@apfelboymchen.net>
 *
 * Zufall is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Zufall is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Juggling of undocumented Xrandr function for multiple monitors via Xrandr
 *
 * https://cgit.freedesktop.org/xorg/proto/randrproto/tree/randrproto.txt
 * https://www.opengl.org/wiki/Programming_OpenGL_in_Linux:_Changing_the_Screen_Resolution
 *
 * CRTC: cathode ray tube controller??
 *
 * useless:
 * for (i = 0; i < xrr_nsize; i++) {
 * 	printf("%2d\t : %-12s: %4i x %4i   (%4imm x%4imm )\n", i, xrr_res->modes[i].name, xrr_sizes[i].width, xrr_sizes[i].height, xrr_sizes[i].mwidth, xrr_sizes[i].mheight);
 * }
 *
 * XRROutputInfo
 *   contains RRMode *modes;
 *   contains int nmode;
 *   holen mit XRRGetOutputInfo()
 * 
 * XRRModeInfo
 *   contains RRMode id;
 *   contains char *name;
 *   contains XRRModeFlags modeFlags; (typedef unsigned long XRRModeFlags;)
 *   contains RRCrtc crtc;
 * 
 * XRRCrtcInfo
 *   contains RRMode mode;
 * 
 * RCrtcInfo *XRRGetCrtcInfo (Display *dpy, XRRScreenResources *resources, RRCrtc crtc);
 * 
 *
 */
#include <stdio.h>

#include <X11/extensions/Xrandr.h>

#include "memman.h"
#include "display-x11-heads.h"
#include "display-x11-xrandr.h"
#include "display-x11-xinerama.h"

static XRRCrtcInfo *mode_used(Display *display, XRRScreenResources *res, RRCrtc crtc, XRRModeInfo *search)
{

	XRRCrtcInfo *crtc_info = XRRGetCrtcInfo(display, res, crtc);
	if (search->id == crtc_info->mode) {
		return crtc_info;
	}
	XRRFreeCrtcInfo(crtc_info);
	return NULL;
}

/*
 * Interface
 */

head_t *xrandr_init(Display *display, int scrnum, Screen *scr, Window root, int *headcount)
{
	head_t *heads = NULL;
	XRRScreenConfiguration *xrr_scrcfg;
	XRRScreenResources *xrr_res;
	XRROutputInfo *xrr_oinfo;

	int infoslots = 1;


	int event_base, error_base;
	if (!XRRQueryExtension (display, &event_base, &error_base)) {
		fprintf(stderr, "RandR extension missing\n");
		return NULL;
	}

	
	heads = getmem(sizeof(head_t));
	heads[0].offset_x = 0;
	heads[0].offset_y = 0;
	heads[0].width    = DisplayWidth(display, scrnum);
	heads[0].height   = DisplayHeight(display, scrnum);

	xrr_scrcfg = XRRGetScreenInfo (display, root);

	xrr_res = XRRGetScreenResources(display, root);
	int output;
	for (output = 0; output < xrr_res->noutput; output++) {
		/* Output */
		xrr_oinfo = XRRGetOutputInfo(display, xrr_res, xrr_res->outputs[output]);
		if (xrr_oinfo->connection != RR_Connected) {
			XRRFreeOutputInfo(xrr_oinfo);
			continue;
		}
		//printf("----- output %d: %s (%d) ", output, xrr_oinfo->name, xrr_oinfo->connection);
		int i;
		for (i = 0; i < xrr_res->nmode; i++) {
			/* modes that can be used on output */
			XRRModeInfo *mode = &xrr_res->modes[i];
			XRRCrtcInfo *crtc_info;
			if ((crtc_info = mode_used(display, xrr_res, xrr_oinfo->crtc, mode) )) {
				/* CRTC found via mode actually used */
				//printf("%s: %d,%d %dx%d -----\n", mode->name, crtc_info->x, crtc_info->y, crtc_info->width, crtc_info->height);
				infoslots++;
				heads = extendmem(heads, sizeof(head_t) * infoslots);
				int slot = infoslots - 1;
				heads[slot].offset_x = crtc_info->x;
				heads[slot].offset_y = crtc_info->y;
				heads[slot].width    = crtc_info->width;
				heads[slot].height   = crtc_info->height;
				XRRFreeCrtcInfo(crtc_info);
				break;
			}
		}
		XRRFreeOutputInfo(xrr_oinfo);
	}
	XRRFreeScreenResources(xrr_res);
	XRRFreeScreenConfigInfo(xrr_scrcfg);

	*headcount = infoslots - 1;
	return heads;
}

/*
 * XRRSelectInput, passing the display, the root window, and the RRScreenChangeNotifyMask mask.
 *
 * Clients must call back into Xlib using XRRUpdateConfiguration when screen 
 * configuration  change  notify events  are generated (or root window 
 * configuration changes occur, to update Xlib's view of the resolution,
 * size, rotation, reflection or subpixel order.
 * -> XSelectInput (dpy, root, StructureNotifyMask);
 */
head_t *xrandr_update(Display *display, int scrnum, Screen *scr, Window root, int *headcount, XEvent *ev)
{
	XRRUpdateConfiguration(ev);
	return xrandr_init(display, scrnum, scr, root, headcount);
}
