/*
 * Copyright (C) 2008, 2009, 2016 Moritz Orbach <zufall@apfelboymchen.net>
 *
 * Zufall is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Zufall is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Parts of this file are ripped from Esetroot.c written by Nat Friedman 
 * <ndf@mit.edu> with modifications by Gerald Britton <gbritton@mit.edu> and 
 * Michael Jennings <mej@eterm.org>
 *
 * management of the X server connection
 * exit(3) on serious programming errors
 *
 * http://www.remlab.net/op/xlib.shtml
 *
 * X11 know-how:
 * https://www.x.org/docs/X11/xlib.pdf
 * http://apfelboymchen.net/mirrors/programming/X-programming/xlib-manual/xlib/window-information/properties-and-atoms.html
 * More explanations below.
 */
#include <stdio.h>
#include <time.h>
#include <string.h>

/*
 * select man page:
 * According to earlier standards
 *     #include <sys/time.h>
 *     #include <sys/types.h>
 *     #include <unistd.h>
 */
#include <sys/select.h>

/* X11 */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xrandr.h>

#include "memman.h"
#include "display.h"
#include "imglib-display-x11.h"
#include "zerror.h"

/* heads stuff */
#include "display-x11-heads.h"

#include "../config.h"
#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
#include "display-x11-xrandr.h"
#endif

#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
#include "display-x11-xinerama.h"
#endif

#include "display-x11-x11.h"

/*
#include "display-x11-xrandr.h"
*/

struct display_s {
	Display *dsp;
	int scrnum;
	Screen *scr;
	Window root;
	Window namewindow;

	int headcount;
	head_t *heads; /* head 0 is the complete surface */

	Pixmap canvas;
	Atom own_XROOTPMAP_ID;
	char *errorstr;
}; /* display_t */

/*
 * stolen from xcompmgr
 */
static void give_me_a_name(display_t *display, char *windowname)
{
	display->namewindow = XCreateSimpleWindow(display->dsp, display->root , 0, 0, 1, 1, 0, None, None);
	Xutf8SetWMProperties(display->dsp, display->namewindow, windowname, windowname, NULL, 0, NULL, NULL, NULL);
}

/*
 * - Property   A collection of named, typed data, identified by an ISO Latin-1 
 *              string
 * - Atom       unique identifier of a property. For efficiency reasons, an 
 *              atom is used rather than a character string.
 *
 * A client can store arbitrary data in a property of a window.
 * The name, type, and value of a property are strings; more precisely, they 
 * are atoms, that is, strings stored in the server and accessible to the 
 * clients via identifiers. A
 *
 * helpful commands for debugging:
 * Xnest -ac -geometry 640x480 :1 &
 * DISPLAY=:1
 * export _Xdebug=1
 * xlsatoms -name _XROOTPMAP_ID -name ESETROOT_PMAP_ID 2>/dev/null
 * xprop -root
 * xprop -root -spy -root
 * xprop -root -remove ESETROOT_PMAP_ID
 *
 * while ./zufall /public/icons/tp.xpm           && test ! -f /tmp/fin;do xprop -root; sleep 0.25 ;done;echo $? > /tmp/fin
 * while ./zufall /public/icons/fvwm.xpm         && test ! -f /tmp/fin;do xprop -root; sleep 0.25 ;done;echo $? > /tmp/fin
 * while ./zufall /public/icons/zoom_img_out.png && test ! -f /tmp/fin;do xprop -root; sleep 0.25 ;done;echo $? > /tmp/fin
 * Esetroot /public/bilder/yosemite\ -\ akw.jpg
 *
 * hm. even Esetroot fails this test.
 *
 * http://www.cs.columbia.edu/~ricardo/misc/Eterm_reference.html
 * http://mail.gnome.org/archives/wm-spec-list/2002-January/msg00006.html
 * http://www.eterm.org/docs/view.php?doc=ref#trans
 * http://foo-projects.org/pipermail/xfce4-dev/2003-January/001699.html
 * 
 * In summary:
 * _XROOTPMAP_ID is the pixmap to be displayed. ESETROOT_PMAP_ID is the
 * pixmap to be destroyed, if any. A program that quits after setting the
 * backdrop will set ESETROOT_PMAP_ID, otherwise it won't, so both types of
 * program are supported.
 *
 * ESETROOT_PMAP_ID is not deleted
 */

static Pixmap get_pixmap_by_atom(display_t *display, Atom *atom)
{
	Pixmap p;
	Atom type = None;
	int format;
	unsigned long length, after;
	unsigned char *data;
	int status;

	/* XA_ATOM instead of AnyPropertyType? Seems more robust to check later. */
	status = XGetWindowProperty(display->dsp, display->root, *atom,
				    0L, 1L, False, AnyPropertyType,
				    &type, &format, &length, &after, &data
				   );

	if (status == Success && type == XA_PIXMAP) {
		/* pointer to pointer to pixmap */
		p = *((Pixmap *)data);
		XFree(data); /* Frees the return of XGetWindowProperty which *points* to the Pixmap */
		return p;
	} else {
		*atom = None;
		return None;
	}
}
static Pixmap get_pixmap_by_property(display_t *display, char *propertyname, Atom *atom)
{
	Atom a; /* storage for when atom == NULL */
	Atom *pa = NULL;
	Pixmap p;

	if (atom == NULL)
		pa = &a;
	else
		pa = atom;

	*pa = XInternAtom(display->dsp, propertyname, True);
	if (*pa == None)
		/* atom does not exist */
		return 0;

	p = get_pixmap_by_atom(display, pa);

	//printf("%s/%d\t0x%06X\n", propertyname, (unsigned int)*pa, (unsigned char) p); // debug

	return p;
}

static int still_set(display_t *display)
{
	int rc;
	Pixmap pixmap_xrootmap = get_pixmap_by_property(display, "_XROOTPMAP_ID", NULL);

	if (pixmap_xrootmap != None) {
		rc = pixmap_xrootmap == display->canvas;
	} else
		rc = -1;

	return rc;

	//printf("\e[1;33mmein 0x%06X vs. 0x%06X\e[0m\n", (unsigned int)display->canvas, (unsigned int)pixmap_xrootmap); // debug

}

/*
 * sets the pixmap for permanent running of zufall.
 * The server must be grabbed already.
 */
static int set_pixmap_running(display_t *display, Pixmap *p)
{
	//printf("set 0x%06X running\n", (unsigned char) *p); // debug
	/* get/create property containing the current background */
	Atom prop_xrootmap = XInternAtom(display->dsp, "_XROOTPMAP_ID", False);
	if (prop_xrootmap == None) {
		errorstr(&display->errorstr, "creation of pixmap property failed.\n");
		return 0;
	}
	display->own_XROOTPMAP_ID = prop_xrootmap;

	/* set prop_xrootmap to the new pixmap */
	XChangeProperty(display->dsp, display->root, prop_xrootmap, XA_PIXMAP, 32, PropModeReplace, (unsigned char *) p, 1);

	return 1;
}

/*
 * sets the pixmap so that zufall can exit (ESETROOT_PMAP_ID to _XROOTPMAP_ID)
 *
 * TODO: check sparen weil gegrabbed?
 */
static int set_pixmap_exiting(display_t *display)
{
	Atom prop_esetroot;

	if (display->canvas == None) {
		/* no image set */
		return 0;
	}
	
	/* ----- GRAB ------ */
	XGrabServer(display->dsp);
	
	/* pointer to pointer to pixmap */
	int ss = still_set(display);
	if (ss == -1) {
		/* error */
		XUngrabServer(display->dsp);
		/* ----- BARG ------ */
		return 1;
	}
	if (!ss) {
		/*
		 * zufall's pixmap is not shown anymore! in the meanwhile 
		 * another program set the background. just clean up the own 
		 * pixmap and return.
		 */
		//printf("\e[1;35mnicht mehr mein background!\e[0m\n"); // debug
		XFreePixmap(display->dsp, display->canvas);
		XUngrabServer(display->dsp);
		/* ----- BARG ------ */
		return 1;
	}

	/* still own pixmap -- set permanent */

	//printf("\t\tstill mine\n"); // debug

	/* get / create */
	prop_esetroot = XInternAtom(display->dsp, "ESETROOT_PMAP_ID", False);
	if (prop_esetroot == None) {
		/* huh? */
		/* fprintf(stderr, "%s: Could not create property\n", PACKAGE_NAME); */
		return 0;
	}

	/*
	 * Setting ESETROOT_PMAP_ID to _XROOTPMAP_ID tells the next program 
	 * that it may kill the pixmap.
	 * From here on, if retain is set, it could be that zufall is killed by 
	 * another process cleaning up the "old" pixmap
	 */
	XChangeProperty(display->dsp, display->root, prop_esetroot, XA_PIXMAP, 32, PropModeReplace, (unsigned char *) &display->canvas, 1);
	/* leave pixmap in X-server after exit. */
	XSetCloseDownMode(display->dsp, RetainPermanent);

	XSync(display->dsp, False); /* necessary? */
	XUngrabServer(display->dsp);
	/* ----- BARG ------ */

	return 1;
}

/*
 * killing the pixmap of a temporary client
 */
static void free_old_persistent(display_t *display)
{
	Atom atom_esetroot;

	Pixmap pixmap_xrootmap = get_pixmap_by_property(display, "_XROOTPMAP_ID", NULL);
	Pixmap pixmap_esetroot = get_pixmap_by_property(display, "ESETROOT_PMAP_ID", &atom_esetroot);
	//printf("0x%06X vs. 0x%06X\n", (unsigned int) pixmap_xrootmap, (unsigned int) pixmap_esetroot); // debug
	if (pixmap_esetroot != None && pixmap_esetroot != None && pixmap_xrootmap == pixmap_esetroot) {
		/*
		 * in the meantime another program set the background and exited
		 * this is the last chance to kill the previous pixmap before 
		 * the reference is overwritten
		 */
		//printf("KILLING 0x%06X\n", (unsigned int) pixmap_esetroot); // debug
		XKillClient(display->dsp, pixmap_esetroot);
		XDeleteProperty(display->dsp, display->root, atom_esetroot);
	}
}


/*
 * pixmap management
 */

/* 
 * render Pixmap pixnew on top of pixbase at given coordinates and size. The 
 * uncovered part of pixbase is filled with hexcolor.
 */
static void render_pixmap(display_t *display, Pixmap *pixbase, Pixmap *pixadd,
		          int head, char *hexcolor, int offset_x, int offset_y, int w, int h)
{
	GC gc;
	XGCValues gcv;
	XColor xcolor;
	Colormap cmap;

	printf("display: h%d %d,%d %dx%d\n", 
			head,
			display->heads[head].offset_x,
			display->heads[head].offset_y,
			display->heads[head].width,
			display->heads[head].height);
	/* create color */
	cmap = DefaultColormap(display->dsp, display->scrnum);
	XParseColor(display->dsp, cmap, hexcolor, &xcolor);
	XAllocColor(display->dsp, cmap, &xcolor);
	gcv.foreground = gcv.background = xcolor.pixel;

	gc = XCreateGC(display->dsp, *pixbase, (GCForeground|GCBackground), &gcv);
	/* fill background with hexcolor */
	XFillRectangle(display->dsp, *pixbase, gc,
		       display->heads[head].offset_x,
		       display->heads[head].offset_y,
		       display->heads[head].width,
		       display->heads[head].height
		      );

	/* sets the tile/stipple origin in the specified GC */
	XSetTSOrigin(display->dsp, gc, display->heads[head].offset_x + offset_x, display->heads[head].offset_y + offset_y);
	XSetFillStyle(display->dsp, gc, FillTiled);
	XSetTile(display->dsp, gc, *pixadd);
	/* image */
	XFillRectangle(display->dsp, *pixbase, gc, display->heads[head].offset_x + offset_x, display->heads[head].offset_y + offset_y, w, h);

	/*
	 * When the grapics context is not freed, the memory for the whole 
	 * pixmap will never be freed by the X server, even when XFreePixmap is called!
	 * This is the "when there are no references to it" part i guess.
	 * This... took... a... while...
	 */
	XFreeGC(display->dsp, gc);
}

/*
 * copy canvas to configured window (most likely root)
 */
static int show_canvas(display_t *display)
{
	int ok;
	/* ----- GRAB ------ */
	XGrabServer(display->dsp);

	free_old_persistent(display);
	ok = set_pixmap_running(display, &display->canvas);

	/* set the background pixmap of the window to the specified pixmap */
	XSetWindowBackgroundPixmap(display->dsp, display->root, display->canvas);

	/* "clear" the window with the background */
	XClearWindow(display->dsp, display->root);

	/* send the request to the server. This is neccessary to update the root-window
	 * True to discard the own changes of the property. Otherwise 
	 * display_sleep will have changes pending on a SIGUSR1 later!
	 */
	XSync(display->dsp, True);
	XUngrabServer(display->dsp);
	/* ----- BARG ------ */

	return ok;
}

static void free_canvas(display_t *display)
{
	if (display->canvas != None) {
		XFreePixmap(display->dsp, display->canvas);
		display->canvas = None;
	}
}

/*
 * ugly replacement functions for all the X extensions managing the screens
 */
head_t *heads_init(Display *display, int scrnum, Screen *scr, Window root, int *headcount)
{
	head_t *head;

#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
	head = xrandr_init(display, scrnum, scr, root, headcount);
	if (head != NULL)
		return head;
#endif
#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
	head = xinerama_init(display, scrnum, scr, headcount);
	if (head != NULL)
		return head;
#endif
	head = headsx11_init(display, scrnum, scr, headcount);
	return head;
}

/* */
head_t *heads_update(Display *display, int scrnum, Screen *scr, Window root, int *headcount, XEvent *ev)
{
	head_t *head;

#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
	head = xrandr_update(display, scrnum, scr, root, headcount, ev);
	return head;
#endif
#ifdef HAVE_X11_EXTENSIONS_XINERAMA_H
	head = xinerama_update(display, scrnum, scr, headcount, ev);
	return head;
#endif
	head = headsx11_update(display, scrnum, scr, headcount, ev);
	return head;
}

/*********************+
 * interface
 ***********************/

/*
 * info
 */

char const * const display_errorstr(display_t *display)
{
	return display->errorstr;
}

int display_headcount(display_t *display)
{
	return display->headcount;
}

void display_headinfo(display_t *display, int head, int *x, int *y, int *w, int *h)
{
	if (head > display->headcount +1) {
		fprintf(stderr, "Programming error: requested head %d, but there are only %d heads\n", head, display->headcount);
		exit(3);
	}

	*x = display->heads[head].offset_x;
	*y = display->heads[head].offset_y;
	*w = display->heads[head].width;
	*h = display->heads[head].height;
}

/*
 * imglib
 */

/*
 * copies the imglib pixmap to the target pixmap respecting background color, 
 * size and location
 */
int display_renderimg(display_t *display, int head, imglib_t *imglib,
		      char *hexcolor, int img_x, int img_y, int img_w, int img_h)
{
	if (head > display->headcount +1) {
		fprintf(stderr, "Programming error: requested head %d, but there are only %d heads\n", head, display->headcount);
		exit(3);
	}

	//printf("head%d: %d,%d %dx%d (%s)\n", head, img_x, img_y, img_w, img_h, hexcolor); // debug
	if (display->canvas == None) {
		/* create Pixmap in root size (display->heads[0]) to be drawn on */
		int depth = DefaultDepth(display->dsp, display->scrnum);
		display->canvas = XCreatePixmap(display->dsp, display->root,
						display->heads[0].width, display->heads[0].height,
						depth
					       );
	}

	Pixmap *p = imglib_render(imglib, img_w, img_h);
	if (p != NULL && *p != None) {
		render_pixmap(display, &display->canvas, p,
				head,
				hexcolor,
				img_x, img_y, img_w, img_h
			     );
		return 1;
	} else
		return 0;
}

/* set image */
int display_showimgs(display_t *display)
{
	return show_canvas(display);
}


/*
 * display server
 */

/*
 * FINALLY! After searching for over a year and even implementing a funny 
 * workaround using a second X connection in a fork, I found the solution very 
 * easily. It's just a matter of the correct search words!
 *
 * http://lists.debian.org/debian-qa/2001/08/msg00039.html
 * http://www.linuxquestions.org/questions/programming-9/xnextevent-select-409355/?s=23201e820d9c750ebc197e352341ca66
 *
 * This function behaves like sleep, but is not only interrupted by signals, 
 * but also by X events.
 *
 * returns
 * - the remaining time to sleep
 */
time_t display_sleep(display_t *display, time_t t, int *repaint)
{
	int x11_fd;
	fd_set select_set;
	time_t start;
	struct timeval tv;
	*repaint = 0;

	x11_fd = ConnectionNumber(display->dsp);

	FD_ZERO(&select_set);
	FD_SET(x11_fd, &select_set);
	tv.tv_sec = t;
	tv.tv_usec = 0;

	start = time(NULL);

	/*
	 * needed, X blocks very unkillable otherwise. TODO: understand why. 
	 *  - it's not the XSelectInput
	 */
	XSync(display->dsp, False);

	/* error detection? A signal is "Interrupted system call"... */
	int rc = select(x11_fd + 1, &select_set, NULL, NULL, &tv);
	if (rc == 0) {
		/*
		 * set time to zero manually if the time limit expired.
		 * select is not accurate enough to unconditionally calculate 
		 * it. The result could be negative or >0
		 */
		t = 0;
	} else {
		/* remaining time */
		t = t - (time(NULL) - start);

		/* t can be < 0 if the program slept longer than it wanted (e.g. busy system) */
		if (t < 0) {
			t = 0;
		}
	}

	/*
	 * Free the pixmap if the _XROOTPMAP_ID was touched by another program
	 *
	 * Detecting wether it was a PropertyChange or not is not possible with 
	 * FD_ISSET, because in the cases of event *and* signal the fd will not 
	 * be removed from the set.
	 *
	 */
	while (XPending(display->dsp)) {
		XEvent ev;

		XNextEvent(display->dsp, &ev);
		//printf("X event: %d\n", ev.type); // debug
		switch (ev.type) {
			case PropertyNotify: /* 28 */
				if (display->canvas != None) {
					XPropertyEvent *event = (XPropertyEvent *) &ev;
					/*
					char *name;
					name = XGetAtomName(event->display, event->atom);
					if (strcmp(name, "_XROOTPMAP_ID") == 0) {
					XFree(name);
					*/
					if (event->atom == display->own_XROOTPMAP_ID) {
						/* property change */
						/* somebody else set a new background. Free own. */
						//printf("root snatched by other program\n"); // debug
						display->own_XROOTPMAP_ID = None;
						free_canvas(display);
					}
				}
				break;
			case RRScreenChangeNotify:
				//printf("X event: RRScreenChangeNotify\n"); // debug
				break;
			case ConfigureNotify:
				//printf("X event: ConfigureNotify\n"); // debug
				free_canvas(display); /* the size of the pixmap changes too */
				freemem(display->heads);
				display->heads = heads_update(display->dsp, display->scrnum, display->scr, display->root, &display->headcount, &ev);
				*repaint = 1;
				break;
			default:
				//printf("X event: wtf? %d (/usr/include/X11/X.h)\n", ev.type); // debug
		}
	}
	/* else: sleep finished or signal */

	return t;
}

/*
 * make sure that the display server is still alive
 */
void display_poke(display_t *display)
{
	/*
	 * XSync will directly call exit() in case the connection does 
	 * not exist anymore
	 */
	XSync(display->dsp, False);
}

void display_close(display_t *display)
{
	if (display == NULL)
		return;
	if (display->errorstr != NULL)
		freemem(display->errorstr);
	if (display->canvas != None)
		set_pixmap_exiting(display);
	XDestroyWindow(display->dsp, display->namewindow);
	XCloseDisplay(display->dsp);
	freemem(display->heads);
	freemem(display);
}

/*
 * open connection to the X server, fill display_t and other 
 * initializations.
 */
display_t *display_open(char *windowname, imglib_t *imglib)
{
	display_t *display;
	display = getmem(sizeof(display_t));

	display->errorstr = getmem(9);
	strcpy(display->errorstr, "no error");

	/* clients without windows don't appear in xlsclients */
	display->dsp = XOpenDisplay(NULL);
	if (!display->dsp) {
		return NULL;
	}
	XSynchronize(display->dsp, True); // debug
	display->scrnum = DefaultScreen(display->dsp);
	display->root = RootWindow(display->dsp, display->scrnum);
	display->scr = ScreenOfDisplay(display->dsp, display->scrnum);
	display->canvas = None;
	give_me_a_name(display, windowname);

	/* receive atom- and resolution changes from the server */
	XSelectInput(display->dsp, display->root, PropertyChangeMask|StructureNotifyMask);

	display->heads = heads_init(display->dsp, display->scrnum, display->scr, display->root, &display->headcount);

	if (display->heads != NULL)
		imglib_displaysetup(imglib, display->dsp, display->root);
	else {
		display_close(display);
		display = NULL;
	}
	display->own_XROOTPMAP_ID = None;

	return display;
}

