/*
 * Copyright (C) 2008, 2009, 2106 Moritz Orbach <zufall@apfelboymchen.net>
 *
 * Zufall is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Zufall is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Imlib2 imglib-interface for zufall
 *
 * https://docs.enlightenment.org/api/imlib2/html/
 * http://adesklets.sourceforge.net/doc/imlib2/imlib2_8c.html
 */
#include <stdio.h>
#include <Imlib2.h>

#include "imglib-display-x11.h"
#include "memman.h"

struct imglib_s {
	Imlib_Image *imlibimg;
	Pixmap pixmap;
	Pixmap mask;

	int display_ok;
}; /* imglib_t */

static int initialized(imglib_t *imglib)
{
	if (imglib != NULL)
		return 1;
	else {
		fprintf(stderr, "imglib: try to use imglib uninitialized!\n");
		return 0;
	}
}
static int display_check(imglib_t *imglib)
{
	if (imglib != NULL && imglib->display_ok) {
		return 1;
	} else {
		fprintf(stderr, "imglib: try to use display uninitialized!\n");
		return 0;
	}
}

/*
 * interface
 */

void imglib_get_dimensions(int *w, int *h)
{
	*w = imlib_image_get_width();
	*h = imlib_image_get_height();
}

void imglib_get_color(int x, int y, int *r, int *g, int *b)
{
	Imlib_Color color_return;

	imlib_image_query_pixel(x, y, &color_return);

	*r = color_return.red;
	*g = color_return.green;
	*b = color_return.blue;
}

/* load and set current image to imgpath */
int imglib_load(imglib_t *imglib, const char *imgpath)
{
	if (!initialized(imglib))
		return 0;

	if (imglib->imlibimg != NULL)
		/* free old image */
		imglib_unload(imglib);

	imglib->imlibimg = imlib_load_image_immediately(imgpath);
	if (imglib->imlibimg != NULL) {
		imlib_context_set_image(imglib->imlibimg);
		return 1;
	} else
		return 0;
}

int imglib_unload(imglib_t *imglib)
{
	if (!initialized(imglib))
		return 0;
	/* pixmaps generated by imlib must also be freed with imlib */
	if (imglib->pixmap != None) {
		imlib_free_image();
		imlib_free_pixmap_and_mask(imglib->pixmap);
	}
	imglib->imlibimg = NULL;
	imglib->pixmap = None;
	imglib->mask = None;

	return 1;
}

int imglib_crop(imglib_t *imglib, int crop_x, int crop_y, int crop_w, int crop_h)
{
	if (!initialized(imglib))
		return 0;

	//printf("imglib crop: %d,%d %d/%d\n", crop_x, crop_y, crop_w, crop_h); // debug
	Imlib_Image *cropped = imlib_create_cropped_image(crop_x, crop_y, crop_w, crop_h);
	imlib_free_image();
	imglib->imlibimg = cropped;
	imlib_context_set_image(imglib->imlibimg);
	return 1;
}

/* render a scaled version of the current image into pixmap_return */
Pixmap *imglib_render(imglib_t *imglib, int w, int h)
{
	if (!initialized(imglib))
		return None;
	if  (!display_check(imglib))
		return None;

	//printf("imglib_render image at size %dx%d\n", w, h); // debug
	imlib_render_pixmaps_for_whole_image_at_size(&imglib->pixmap, &imglib->mask, w, h);
	return &imglib->pixmap;
}

int imglib_displaysetup(imglib_t *imglib, Display *dsp, Window targetwin)
{
	if (!initialized(imglib))
		return 0;
	/* target for imlib_render_pixmaps_for_whole_image_at_size() */
	int ds = DefaultScreen(dsp);
	Visual *vis = DefaultVisual(dsp, ds);
	Colormap cm = DefaultColormap(dsp, ds);
	imlib_context_set_display(dsp);
	imlib_context_set_visual(vis);
	imlib_context_set_colormap(cm);
	imlib_context_set_drawable(targetwin);
	imlib_context_set_blend(0);
	imglib->display_ok = 1;

	return 1;
}

imglib_t *imglib_init(void)
{
	imglib_t *imglib;
	imglib = getmem(sizeof(imglib_t));

	imglib->imlibimg = NULL;
	imglib->pixmap = None;
	imglib->mask = None;
	imglib->display_ok = 0;
	imlib_set_cache_size(0); /* caching eats up memory and is useless in this type of program */

	return imglib;
}

void imglib_destroy(imglib_t *imglib)
{
	if (imglib == NULL)
		return;

	freemem(imglib);
}

