/*
 * Copyright (C) 2004, 2005, 2006, 2007, 2016
 * Moritz Orbach <zufall@apfelboymchen.net>
 *
 * Zufall is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Zufall is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Traverses the file system and adds files to a linked list.
 *
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* files and dirs */
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include "../config.h"
#include "recurse.h"
#include "zerror.h"

#ifdef HAVE_REALPATH
#include <stdlib.h>
#include "list.h"

/*
 *
 */

list_t *visited = NULL;
int status_had_dirs;
#endif


#ifdef HAVE_REALPATH
/* symlink-loop-detection */

int alreadythere(list_t *visited, const char *path)
{
	int i;
	char *savedpath = NULL;

	for (i = 1; i <= list_count(visited); i++) {
		savedpath = list_get(visited, i);
		//printf("%3d: %s vs. %s\n", i, path, savedpath); // debug
		if (!strcmp(path, savedpath))
			return 1;
	}
	return 0;
}

int check_loop(list_t *visited, const char *dir)
{
	char *rp; /* realpath */
	int usable;

	rp = realpath(dir, NULL);
	if (rp != NULL) {
		if (!alreadythere(visited, rp)) {
			//printf("not there: %s\n\n", rp); // debug
			list_append(visited, rp, strlen(rp) + 1);
			usable = 1;
		} else {
			usable = 0;
		}
		free(rp);
	} else {
		/* strange ... */
		usable = 0;
		printerrno(dir);
	}
	return usable;
}
#endif

int had_dirs(void)
{
	return status_had_dirs;
}

/*
 * adds a file / the files in the directory recursivly to filesearch in *config
 *
 * *sb is used to save stats and may be NULL.
 */
static int addpath(filesearch_t *config, char const *const path, struct stat *sb)
{
	struct dirent *dentry;
	struct stat lsb; /* local sb */
	static int fatal_error = 0; /* function will return if this is true */

	if (sb == NULL) {
		if (stat(path, &lsb)) {
			printerrno(path);
			return 0;
		}
	} else
		lsb = *sb;

	if (! S_ISDIR(lsb.st_mode)) {
		/* path is a FILE or link to a file - add it to the list */
		filesearch_add(config, path, &lsb);
	} else {
		/* path is a DIRECTORY or a link to a directory - read it */
		int noloop = 1;
#ifdef HAVE_REALPATH
		noloop = check_loop(visited, path);
#endif
		if (noloop) {
			DIR* dirp;
			dirp = opendir(path);
			if (dirp == NULL) {
				/* error or file? */
				printerrno(path);
				return 0;
			}

			/* reading directory entries */
			while (errno = 0, (dentry = readdir(dirp)) != NULL) { /* readdir doesn't set errno itself */
				/* FILES within a directory again */
				if (dentry->d_name[0] == '.')
					continue;

				/* make full path */
				char fullpath[strlen(path) + strlen(dentry->d_name) + 2]; /* +1 \0, +1 '/'*/
				strcpy(fullpath, path);
				/* append slash if there is none (every time, I guess) */
				if (fullpath[strlen(fullpath) - 1] != '/')
					strncat(fullpath, "/", 1);
				strcat(fullpath, dentry->d_name);

				if (stat(fullpath, &lsb) == -1) {
					if (errno == ENOENT)
						fprintf(stderr, "%s: dead symlink: %s\n", PACKAGE, fullpath);
					else
						printerrno(fullpath);
					continue;
				}
				addpath(config, fullpath, &lsb);
				if (fatal_error)
					break;
			}
			if (errno != 0)
				printerrno(path);
			closedir(dirp);
		} else {
			fprintf(stderr, "%s: symlink loop detected: %s\n", PACKAGE , path);
		}
	}

	return fatal_error == 0;
}

/*
 * TODO canonical
 * 	const?
 */
int recurse(filesearch_t *config, int pathc, char *pathv[])
{
#ifdef HAVE_REALPATH
	visited = list_init();
#endif

	int ok = 1;
	int i;

	status_had_dirs = 0;
	for (i = 0; ok && i < pathc; i++) {
		struct stat sb;
		if (stat(pathv[i], &sb)) {
			printerrno(pathv[i]);
			return 0;
		}
		if (S_ISDIR(sb.st_mode)) {
			status_had_dirs = 1; /* zufall wants to know this */
		}

		ok = addpath(config, pathv[i], &sb);
	}

#ifdef HAVE_REALPATH
	list_destroy(visited);
#endif

	return ok == 1;
}
