/*
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2016
 * Moritz Orbach <zufall@apfelboymchen.net>
 * http://apfelboymchen.net/gnu/C/zufall/
 *
 * Zufall is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Zufall is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Main program
 */
#include <stdio.h>
#include <stdlib.h>

#include <time.h>
#include <signal.h>
#include <string.h>

/* getpid */
#include <sys/types.h>
#include <unistd.h>

/* internals */
#include "../config.h"
#include "cfg.h"
#include "list.h"
#include "filesearch.h"
#include "memman.h"
#include "status.h"
#include "display.h"
#include "imglib.h"
#include "bgcalc.h"

/* file gathering */
#include "recurse.h"
#include "collection.h"
#include "command.h"

static char *HARDCFG_supported_filetypes[] = {
	"jpg",
	"jpeg",
	"png",
	"gif",
	"bmp",
	"tga",
	"tif",
	"tiff",
	NULL
};

typedef struct flow_s {
	int head_from;
	int head_to;
	int image_per_screen;
	unsigned int zufall; /* index of image in *imglist */
	unsigned int tr;
} flow_t;
status_t *status; /* this has to be global because it is (illegally) used in the signal handlers */

/* flow */
void flow_init(flow_t *flow, int headcount, unsigned int zufall)
{
	flow->zufall = zufall;
	/*
	 * Decide for which and on how many screens the image(s) must be set
	 * - pseudo-head 0 is the whole area of alle screens together
	 * - display_headcount() returns only real screens
	 */
	switch (cfg.mh_mode) {
		case STRETCH:
			/* pseudo-head */
			flow->head_from = 0;
			flow->head_to = 0;
			break;
		case CLONE:
		case PER_SCREEN:
		default:
			/* nothing special - one pass per monitor */
			flow->head_from = 1;
			flow->head_to = headcount;
			break;
	}
	if (cfg.mh_mode != CLONE)
		flow->image_per_screen = 1; /* one image per screen */
	else
		flow->image_per_screen = 0; /* same image on all screens (-M stretch) */
}


/*
 * signal handlers
 */
void signal_quit(int p)
{
	status_stop(status);
	status_break_sleep(status, EXIT);
}
void signal_next(int p)
{
	status_break_sleep(status, NEXTIMG);
}
void infotrap(int p)
{
	/* calling signal() (POSIX allows it to do within the signal handler) 
	 * shouldn't be needed on nodern systems */
	//signal(SIGUSR1, infotrap);
	status_break_sleep(status, INFO);
}

unsigned int getrandom(unsigned int max)
{
	if (max > 0)
		return(rand() % max + 1);
	else
		return 0;
}

/*************************
 * functionality
 *************************/

int option_command(status_t *status, command_t *command, char *imgpath)
{
	int w, h;

	status_recall_size(status, &w, &h);
	return command_execute(command, status_recall_color(status), imgpath, w, h);
}

/*
 * - calculates size and colors of the image to be displayed
 * - renders *imgfilename in the display object
 */
void calculate_bg(status_t *status, display_t *display, int headnum)
{
	int img_w, img_h, img_x, img_y;
	color_t c1, c2;
	int head_x, head_y, head_w, head_h;

	char hexcolor[8];
	imglib_get_dimensions(&img_w, &img_h); 
	imglib_get_color(0, 0, &c1.red, &c1.green, &c1.blue);
	imglib_get_color(0, img_h-1, &c2.red, &c2.green, &c2.blue);
	nice_color(hexcolor, c1, c2);


	/* calculate position (centered) and image dimensions */
	display_headinfo(display, headnum, &head_x, &head_y, &head_w, &head_h);
	int crop_w, crop_h;
	image_scaledsize(head_w, head_h, &img_w, &img_h, &crop_w, &crop_h, cfg.maxscale, cfg.minborder, cfg.crop);

	status_remember_cropping(status, 0, 0, crop_w, crop_h);

	/* center */
	img_x = (head_w - img_w) / 2;
	img_y = (head_h - img_h) / 2;

	status_remember_color(status, hexcolor);
	status_remember_offset(status, img_x, img_y);
	status_remember_size(status, img_w, img_h);
}

/*****************+
 * helper
 ******************/

void signalinfo(status_t *status, flow_t flow, time_t tr, list_t *imglist)
{
	int head;
	for (head = 0; head < flow.head_to; head++){
		char *img = status_recall_name(status, head +1); /* lists start at 1 */
		printf("%s: %s\n", PACKAGE, img);
	}
	/* damn, time_t shouldn't be printed! */
	printf("%s: %.1ld:%.2ld remaining, %d image(s)\n", PACKAGE, (long int)tr / 60, (long int)tr % 60, list_count(imglist));
}

/*
 * Wait for signal, display server event or passed delay for the current image.
 *
 * This must be handled by the display server because the sleep must possibly 
 * be interrupted by display server events.
 */
void zsleep(flow_t *flow, status_t *status, display_t *display, list_t *imglist)
{
	int repaint;
	time_t tr;
	int continue_sleep = 1;

	tr = flow->tr;

	status_break_sleep(status, NONE);
	while (continue_sleep && tr != 0) {
		tr = display_sleep(display, tr, &repaint); /* admire background */

		/* sleep is over. Event? Signal? Continue sleep or new sleep? */
		if (repaint)
			status_break_sleep(status, REPAINT);

		//printf("sleep broken: %d\n", status_sleep_broken(status));
		switch (status_sleep_broken(status)) {
			case NEXTIMG:
				status_break_sleep(status, NONE);
				/* no break, rest is like EXIT */
			case EXIT:
				/* done for now */
				continue_sleep = 0;
				tr = 0;
				break;
			case INFO: /* print info and continue sleep */
				signalinfo(status, *flow, tr, imglist);
				status_break_sleep(status, NONE);
				break;
			case REPAINT: /* don not continue sleep, but don't reset tr either */
				continue_sleep = 0;
				flow_init(flow, display_headcount(display), flow->zufall);
				break;
			case NONE: /* do nothing, sleep tr */
				break;
		}
	}
	flow->tr = tr;
}

/*
 * Des Zufalls Schleife
 */

int load_image(status_t *status, display_t *display, int head, imglib_t *imglib, list_t *imglist, unsigned int zufall)
{
	int success;
	char *imgpath;
	imgpath = list_get(imglist, zufall);

	if (imglib_load(imglib, imgpath)) {
		success = 1;
	} else {
		success = 0;
		fprintf(stderr, "%s: unable to load image file \"%s\".\n", PACKAGE, imgpath);
		if (list_delete(imglist, zufall)) {
			if (list_count(imglist) == 0)
				/* deleted last image */
				status_stop(status);
		} else {
			/* list_delete failed?? */
			fprintf(stderr, "error freeing memory..\n");
			success = 0;
			status_stop(status);
		}
	}
	printf("so ein zufall!: %s\n", imgpath); /* yes, the punctuation is a bit extravagant */
	return success;
}

/* render into display server */
int render(status_t *status, display_t *display, int head, imglib_t *imglib)
{
	int x, y, w, h;
	char *color;
	int crop_x, crop_y, crop_w, crop_h;

	color = status_recall_color(status);
	status_recall_offset(status, &x, &y);
	status_recall_size(status, &w, &h);
	status_recall_cropping(status, &crop_x, &crop_y, &crop_w, &crop_h);
	imglib_crop(imglib, crop_x, crop_y, crop_w, crop_h);
	return display_renderimg(display, head, imglib, color, x, y, w, h);
}

/* return new index, possibly based on old index */
unsigned int get_index(list_t *imglist, int head, unsigned int index, int error)
{
	if (cfg.mh_mode == CLONE && head > 1 && index > 0)
		/* return same image if heads are clones and the image is already selected */
		return index;

	if (!cfg.sequential)
		index = getrandom(list_count(imglist));
	else {
		/* the "next" image moves to the last index if an image is deleted */
		if (!error)
			index++;
		if (index > list_count(imglist))
			/* wrap around */
			index = 1; /* this function is not called if there's no image left */
	}

	return index;
}


/*
 * returns 0 if zufall should not continue
 */
int manage_passes(long int images_available, signed long int *passesleft)
{
	int cont = 1;

	if (images_available == 1)
		/* this iteration set the last image */
		cont = 0;
	if (cfg.passes != -1) {
		/* passes are limited */
		*passesleft = *passesleft - 1;
		if (*passesleft == 0)
			/* last pass passed */
			cont = 0;
	}

	return cont;
}

/*
 * - get new image
 * - load image from disk
 * - load image into display
 * returns success/failure
 */
int do_head(int head, flow_t *flow, int error, status_t *status, display_t *display, imglib_t *imglib, list_t *imglist)
{
	char *imgpath;
	int success;

	/* get new image */
	flow->zufall = get_index(imglist, head, flow->zufall, error);

	imgpath = list_get(imglist, flow->zufall);
	if (load_image(status, display, head, imglib, imglist, flow->zufall)) {
		success = 1;
		status_remember_image(status, head, imgpath);
		calculate_bg(status, display, head);
	} else
		success = 0;

	return success;
}

/*
 * - load image from disk based on currently loades images
 * - load image into display
 * returns success/failure
 */
int do_head_repaint(int head, flow_t *flow, int error, status_t *status, display_t *display, imglib_t *imglib, list_t *imglist)
{
	char *imgpath;
	int success = 1;

	imgpath = status_recall_name(status, head);
	if (imglib_load(imglib, imgpath)) {
		success = 1;
		calculate_bg(status, display, head);
	} else
		success = 0;

	return success;
}

/*
 */
int zufall(display_t *display, command_t *command, imglib_t *imglib, list_t *imglist)
{
	signed long int passesleft = cfg.passes;
	int initial_count = list_count(imglist);
	int error = 0; /* stop() is also called on SIGTERM/SIGING */
	flow_t flow;

	flow_init(&flow, display_headcount(display), 0);
	flow.tr = cfg.delay;

	while (status_is_running(status)) {
		int head;
		int repaint = status_sleep_broken(status) == REPAINT;

		if (!repaint)
			status_forget_images(status);
		for (head = flow.head_from; head <= flow.head_to && status_is_running(status); ) {
			/* head */
			if (!repaint)
				error = !do_head(head, &flow, error, status, display, imglib, imglist);
			else
				error = !do_head_repaint(head, &flow, error, status, display, imglib, imglist);

			if (!error && status_is_running(status)) {
				if (!render(status, display, head, imglib)) {
					fprintf(stderr, "%s: failed to prepare background!\n", PACKAGE);
					error = 1;
					status_stop(status);
				}
			}

			if (!error) {
				if (cfg.command != NULL) {
					option_command(status, command, status_recall_name(status, head));
				}
				head++;
			}
		} /* head */
		imglib_unload(imglib);
		if (!repaint)
			flow.tr = cfg.delay;

		/* what to do now? */
		if (status_is_running(status)) {
			if (cfg.setbg) {
				if (!display_showimgs(display)) {
					fprintf(stderr, "%s: failed to set background: %s\n", PACKAGE, display_errorstr(display));
					error = 1;
					status_stop(status);
				}
			}
			//printf("loop: %ld pass(es) left, %d image(s)\n\n", passesleft, list_count(imglist)); // debug
			if (status_is_running(status) && manage_passes(list_count(imglist), &passesleft)) {
				/* only sleep if the loop is still active */
				zsleep(&flow, status, display, imglist);
			} else
				/* last pass passed */
				status_stop(status);
		}
	} /* running */

	/* exit */
	if (initial_count > 1 || had_dirs()) {
		/* don't print anything if there was only one picture in the first place */
		if (status_sleep_broken(status) == EXIT || error)
			printf("%s: exiting\n", PACKAGE);
		else
			printf("%s: no pictures left, exiting\n", PACKAGE);
	}

	return error == 0;
}

/*
 * init
 */

/* 
 * - parses command line
 * - initializations
 *
 * returns:
 * - 0 on successful initialization
 * - 1 if zufall should not run, but no error occurd
 * - 2 on error
 */
int init(int argc, char *argv[], display_t **display, command_t **command, imglib_t **imglib, list_t **imglist)
{
	/*
	 * command line handling
	 */
	int diroffset; /* where the directories start in *argv[] */
	filesearch_t *filesearch = NULL; /* interface for searching files */
	/* geeqie collections or files with one image per line to be read into *imglist */
	char **collections = NULL;
	int collectionsc;

	switch (cfg_cmdln(argc, argv, &diroffset, &collections, &collectionsc)) {
		case 1: return 1; /* --help, --version */
		case 2: return 2; /* error */
		case 0: break;    /* success */
	}

	*imglib = imglib_init();
	if (*imglib == NULL) {
		fprintf(stderr, "%s: unable initialize image interface\n", PACKAGE);
		return 2;
	}

	status = status_init();
	if (status == NULL) {
		fprintf(stderr, "%s: unable initialize status interface\n", PACKAGE);
		return 2;
	}

	/* a display connection is always needed because the display size 
	 * affects the image size in -c */
	*display = display_open(PACKAGE, *imglib);
	if (*display == NULL) {
		fprintf(stderr, "%s: unable to open display\n", PACKAGE);
		return 2;
	}

	if (cfg.command) {
		*command = command_init(cfg.command);
		if (!command_valid(*command)) {
			fprintf(stderr, "%s: %s\n", PACKAGE, command_errorstr(*command));
			return 2;
		}
	}

	/*
	 * get images via filesearch
	 */

	/* building the image list may take a while, so ignore signals used for communication */
	if (signal(SIGHUP,  SIG_IGN) == SIG_ERR ||
	    signal(SIGUSR1, SIG_IGN) == SIG_ERR
	   ) {
		fprintf(stderr, "%s: unable to set up signal handlers\n", PACKAGE);
		return 2;
	}

	*imglist =  NULL; /* let filesearch create it */
	filesearch = filesearch_init(*imglist, HARDCFG_supported_filetypes, cfg.minsize);
	if (filesearch == NULL) {
		fprintf(stderr, "%s: unable to initialize filesearch\n", PACKAGE);
		return 2;
	}

	/* collect files from paths */
	if (!recurse(filesearch, argc - diroffset, &argv[diroffset])) {
		fprintf(stderr, "%s: fatal error while searching for files\n", PACKAGE);
		return 2;
	}

	if (collectionsc > 0) {
		/* collect files from stdin, list files or geeqie collection files */
		collection(filesearch, collectionsc, collections);
		freemem(collections); /* allocated by cfg_cmdln */
	}
	*imglist = filesearch_getlist(filesearch);
	filesearch_destroy(filesearch, 0);

	/*
	 * misc
	 */

	/* kill / CTRL-C. sigaction would be preferred, but is not portable */
	if (signal(SIGTERM, signal_quit) == SIG_ERR ||
	    signal(SIGINT,  signal_quit) == SIG_ERR ||
	    signal(SIGHUP,  signal_next) == SIG_ERR ||
	    signal(SIGUSR1, infotrap)    == SIG_ERR
	   ) {
		fprintf(stderr, "%s: unable to set up signal handlers\n", PACKAGE);
		return 2;
	}

	/* initiate the zufall */
	srand(time(NULL));

	/*
	 * don't buffer stdout more than a line, so that the filename can be 
	 * seen immediately with tail -f when the image changes.
	 * "The standard error stream stderr is always unbuffered by default."
	 */
	if (setvbuf(stdout, (char *)NULL, _IOLBF, 0) != 0) {
		fprintf(stderr, "%s: unable to set up buffering\n", PACKAGE);
		return 2;
	}

	return 0;
}


/*
 * main
 */

int main(int argc, char *argv[])
{

	display_t *display = NULL;       /* display server interface */
	command_t *command = NULL;       /* interface for executing commands */
	imglib_t *imglib = NULL;         /* image library interface */
	list_t *imglist = NULL;
	status = NULL;
	int rc; /* shell return code */

	rc = init(argc, argv, &display, &command, &imglib, &imglist);
	if (rc == 0) {
		if (list_count(imglist) > 0) {
			rc = zufall(display, command, imglib, imglist);
			if (rc != 1)
				rc = 2; // convert to shell
		} else {
			fprintf(stderr, "%s: Sorry, no supported images found\n", PACKAGE);
			rc = 1;
		}
	}
	if (rc == 1)
		rc = 0; /* this is no error */

	int sigint = 0;
	if (status != NULL)
		sigint = status_sleep_broken(status) == EXIT;

	/* clean up */
	status_forget_images(status);
	status_destroy(status);
	list_destroy(imglist);
	imglib_destroy(imglib);
	display_close(display);
	command_destroy(command);

	if (sigint) {
		/* let the parent know
		 * https://www.cons.org/cracauer/sigint.html
		 */
		signal(SIGINT, SIG_DFL);
		kill(getpid(), SIGINT);
	}
	return rc;
}
